/*
 * Default style sheet on Windows Platforms
 */
ads--CDockContainerWidget {
	background: palette(dark);
}
ads--CDockContainerWidget > QSplitter{
        padding: 1 0 1 0;
}

ads--CDockContainerWidget ads--CDockSplitter::handle {
	background: palette(dark);
}

ads--CDockAreaWidget {
	background: palette(window);
	border: 1px solid white;
}

ads--CDockWidgetTab {
	background: palette(window);
	border-color: palette(light);
	border-style: solid;
	border-width: 0 1px 0 0;
	padding: 0 0px;
}

ads--CDockWidgetTab[activeTab="true"] {
	background: qlineargradient(spread : pad, x1 : 0, y1 : 0, x2 : 0, y2 : 0.5, stop : 0
		palette(window), stop:1 palette(light));
	/*background: palette(highlight);*/
}

ads--CDockWidgetTab QLabel {
	color: palette(dark);
}

ads--CDockWidgetTab[activeTab="true"] QLabel {
	color: palette(foreground);
}

ads--CDockWidget {
	background: palette(light);
	border-color: palette(light);
	border-style: solid;
	border-width: 1px 0 0 0;
}

ads--CTitleBarButton {
	padding: 0px 0px;
}

QScrollArea#dockWidgetScrollArea {
	padding: 0px;
	border: none;
}

#tabCloseButton {
	margin-top: 2px;
	background: none;
	border: none;
	padding: 0px -2px;
	qproperty-icon: url(:/ads/images/close-button.svg),
		url(:/ads/images/close-button-disabled.svg) disabled;
	qproperty-iconSize: 16px;
}

#tabCloseButton:hover {
	border: 1px solid rgba(0, 0, 0, 32);
	background: rgba(0, 0, 0, 16);
}

#tabCloseButton:pressed {
	background: rgba(0, 0, 0, 32);
}

#tabsMenuButton::menu-indicator {
	image: none;
}

#tabsMenuButton {
	qproperty-icon: url(:/ads/images/tabs-menu-button.svg);
	qproperty-iconSize: 16px;
}

#dockAreaCloseButton {
	qproperty-icon: url(:/ads/images/close-button.svg),
		url(:/ads/images/close-button-disabled.svg) disabled;
	qproperty-iconSize: 16px;
}

#detachGroupButton {
	qproperty-icon: url(:/ads/images/detach-button.svg),
		url(:/ads/images/detach-button-disabled.svg) disabled;
	qproperty-iconSize: 16px;	
}
